@@ -33,6 +33,7 @@ import (
 
 	dto "github.com/prometheus/client_model/go"
 
+	//nolint:staticcheck // Ignore SA1019. Need to keep deprecated package for compatibility.
 	"github.com/golang/protobuf/proto"
 	"github.com/prometheus/common/expfmt"
 
@@ -53,7 +54,7 @@ func (u uncheckedCollector) Collect(c chan<- prometheus.Metric) {
 func testHandler(t testing.TB) {
 	// TODO(beorn7): This test is a bit too "end-to-end". It tests quite a
 	// few moving parts that are not strongly coupled. They could/should be
-	// tested separately. However, the changes planned for v0.10 will
+	// tested separately. However, the changes planned for v2 will
 	// require a major rework of this test anyway, at which time I will
 	// structure it in a better way.
 
@@ -259,8 +260,9 @@ collected metric "name" { label:<name:"constname" value:"\377" > label:<name:"la
 `)
 
 	summary := prometheus.NewSummary(prometheus.SummaryOpts{
-		Name: "complex",
-		Help: "A metric to check collisions with _sum and _count.",
+		Name:       "complex",
+		Help:       "A metric to check collisions with _sum and _count.",
+		Objectives: map[float64]float64{0.5: 0.05, 0.9: 0.01, 0.99: 0.001},
 	})
 	summaryAsText := []byte(`# HELP complex A metric to check collisions with _sum and _count.
 # TYPE complex summary
@@ -709,12 +711,12 @@ collected metric "broken_metric" { label:<name:"foo" value:"bar" > label:<name:"
 			registry.MustRegister(scenario.collector)
 		}
 		writer := httptest.NewRecorder()
-		handler := prometheus.InstrumentHandler("prometheus", promhttp.HandlerFor(gatherer, promhttp.HandlerOpts{}))
+		handler := promhttp.HandlerFor(gatherer, promhttp.HandlerOpts{})
 		request, _ := http.NewRequest("GET", "/", nil)
 		for key, value := range scenario.headers {
 			request.Header.Add(key, value)
 		}
-		handler(writer, request)
+		handler.ServeHTTP(writer, request)
 
 		for key, value := range scenario.out.headers {
 			if writer.Header().Get(key) != value {
@@ -745,37 +747,137 @@ func BenchmarkHandler(b *testing.B) {
 }
 
 func TestAlreadyRegistered(t *testing.T) {
-	reg := prometheus.NewRegistry()
 	original := prometheus.NewCounterVec(
 		prometheus.CounterOpts{
-			Name: "test",
-			Help: "help",
+			Name:        "test",
+			Help:        "help",
+			ConstLabels: prometheus.Labels{"const": "label"},
 		},
 		[]string{"foo", "bar"},
 	)
 	equalButNotSame := prometheus.NewCounterVec(
+		prometheus.CounterOpts{
+			Name:        "test",
+			Help:        "help",
+			ConstLabels: prometheus.Labels{"const": "label"},
+		},
+		[]string{"foo", "bar"},
+	)
+	originalWithoutConstLabel := prometheus.NewCounterVec(
 		prometheus.CounterOpts{
 			Name: "test",
 			Help: "help",
 		},
 		[]string{"foo", "bar"},
 	)
-	var err error
-	if err = reg.Register(original); err != nil {
-		t.Fatal(err)
+	equalButNotSameWithoutConstLabel := prometheus.NewCounterVec(
+		prometheus.CounterOpts{
+			Name: "test",
+			Help: "help",
+		},
+		[]string{"foo", "bar"},
+	)
+
+	scenarios := []struct {
+		name              string
+		originalCollector prometheus.Collector
+		registerWith      func(prometheus.Registerer) prometheus.Registerer
+		newCollector      prometheus.Collector
+		reRegisterWith    func(prometheus.Registerer) prometheus.Registerer
+	}{
+		{
+			"RegisterNormallyReregisterNormally",
+			original,
+			func(r prometheus.Registerer) prometheus.Registerer { return r },
+			equalButNotSame,
+			func(r prometheus.Registerer) prometheus.Registerer { return r },
+		},
+		{
+			"RegisterNormallyReregisterWrapped",
+			original,
+			func(r prometheus.Registerer) prometheus.Registerer { return r },
+			equalButNotSameWithoutConstLabel,
+			func(r prometheus.Registerer) prometheus.Registerer {
+				return prometheus.WrapRegistererWith(prometheus.Labels{"const": "label"}, r)
+			},
+		},
+		{
+			"RegisterWrappedReregisterWrapped",
+			originalWithoutConstLabel,
+			func(r prometheus.Registerer) prometheus.Registerer {
+				return prometheus.WrapRegistererWith(prometheus.Labels{"const": "label"}, r)
+			},
+			equalButNotSameWithoutConstLabel,
+			func(r prometheus.Registerer) prometheus.Registerer {
+				return prometheus.WrapRegistererWith(prometheus.Labels{"const": "label"}, r)
+			},
+		},
+		{
+			"RegisterWrappedReregisterNormally",
+			originalWithoutConstLabel,
+			func(r prometheus.Registerer) prometheus.Registerer {
+				return prometheus.WrapRegistererWith(prometheus.Labels{"const": "label"}, r)
+			},
+			equalButNotSame,
+			func(r prometheus.Registerer) prometheus.Registerer { return r },
+		},
+		{
+			"RegisterDoublyWrappedReregisterDoublyWrapped",
+			originalWithoutConstLabel,
+			func(r prometheus.Registerer) prometheus.Registerer {
+				return prometheus.WrapRegistererWithPrefix(
+					"wrap_",
+					prometheus.WrapRegistererWith(prometheus.Labels{"const": "label"}, r),
+				)
+			},
+			equalButNotSameWithoutConstLabel,
+			func(r prometheus.Registerer) prometheus.Registerer {
+				return prometheus.WrapRegistererWithPrefix(
+					"wrap_",
+					prometheus.WrapRegistererWith(prometheus.Labels{"const": "label"}, r),
+				)
+			},
+		},
 	}
-	if err = reg.Register(equalButNotSame); err == nil {
-		t.Fatal("expected error when registering equal collector")
+
+	for _, s := range scenarios {
+		t.Run(s.name, func(t *testing.T) {
+			var err error
+			reg := prometheus.NewRegistry()
+			if err = s.registerWith(reg).Register(s.originalCollector); err != nil {
+				t.Fatal(err)
+			}
+			if err = s.reRegisterWith(reg).Register(s.newCollector); err == nil {
+				t.Fatal("expected error when registering new collector")
+			}
+			if are, ok := err.(prometheus.AlreadyRegisteredError); ok {
+				if are.ExistingCollector != s.originalCollector {
+					t.Error("expected original collector but got something else")
+				}
+				if are.ExistingCollector == s.newCollector {
+					t.Error("expected original collector but got new one")
+				}
+			} else {
+				t.Error("unexpected error:", err)
+			}
+		})
 	}
-	if are, ok := err.(prometheus.AlreadyRegisteredError); ok {
-		if are.ExistingCollector != original {
-			t.Error("expected original collector but got something else")
-		}
-		if are.ExistingCollector == equalButNotSame {
-			t.Error("expected original callector but got new one")
-		}
-	} else {
-		t.Error("unexpected error:", err)
+}
+
+// TestRegisterUnregisterCollector ensures registering and unregistering a
+// collector doesn't leave any dangling metrics.
+// We use NewGoCollector as a nice concrete example of a collector with
+// multiple metrics.
+func TestRegisterUnregisterCollector(t *testing.T) {
+	col := prometheus.NewGoCollector()
+
+	reg := prometheus.NewRegistry()
+	reg.MustRegister(col)
+	reg.Unregister(col)
+	if metrics, err := reg.Gather(); err != nil {
+		t.Error("error gathering sample metric")
+	} else if len(metrics) != 0 {
+		t.Error("should have unregistered metric")
 	}
 }
 
@@ -918,6 +1020,11 @@ test_summary_count{name="foo"} 2
 		prometheus.SummaryOpts{
 			Name: "test_summary",
 			Help: "test summary",
+			Objectives: map[float64]float64{
+				0.5:  0.05,
+				0.9:  0.01,
+				0.99: 0.001,
+			},
 		},
 		[]string{"name"},
 	)
@@ -975,6 +1082,96 @@ test_summary_count{name="foo"} 2
 	fileContents := string(fileBytes)
 
 	if fileContents != expectedOut {
-		t.Error("file contents didn't match unexpected")
+		t.Errorf(
+			"files don't match, got:\n%s\nwant:\n%s",
+			fileContents, expectedOut,
+		)
+	}
+}
+
+// collidingCollector is a collection of prometheus.Collectors,
+// and is itself a prometheus.Collector.
+type collidingCollector struct {
+	i    int
+	name string
+
+	a, b, c, d prometheus.Collector
+}
+
+// Describe satisifies part of the prometheus.Collector interface.
+func (m *collidingCollector) Describe(desc chan<- *prometheus.Desc) {
+	m.a.Describe(desc)
+	m.b.Describe(desc)
+	m.c.Describe(desc)
+	m.d.Describe(desc)
+}
+
+// Collect satisifies part of the prometheus.Collector interface.
+func (m *collidingCollector) Collect(metric chan<- prometheus.Metric) {
+	m.a.Collect(metric)
+	m.b.Collect(metric)
+	m.c.Collect(metric)
+	m.d.Collect(metric)
+}
+
+// TestAlreadyRegistered will fail with the old, weaker hash function.  It is
+// taken from https://play.golang.org/p/HpV7YE6LI_4 , authored by @awilliams.
+func TestAlreadyRegisteredCollision(t *testing.T) {
+
+	reg := prometheus.NewRegistry()
+
+	for i := 0; i < 10000; i++ {
+		// A collector should be considered unique if its name and const
+		// label values are unique.
+
+		name := fmt.Sprintf("test-collector-%010d", i)
+
+		collector := collidingCollector{
+			i:    i,
+			name: name,
+
+			a: prometheus.NewCounter(prometheus.CounterOpts{
+				Name: "my_collector_a",
+				ConstLabels: prometheus.Labels{
+					"name": name,
+					"type": "test",
+				},
+			}),
+			b: prometheus.NewCounter(prometheus.CounterOpts{
+				Name: "my_collector_b",
+				ConstLabels: prometheus.Labels{
+					"name": name,
+					"type": "test",
+				},
+			}),
+			c: prometheus.NewCounter(prometheus.CounterOpts{
+				Name: "my_collector_c",
+				ConstLabels: prometheus.Labels{
+					"name": name,
+					"type": "test",
+				},
+			}),
+			d: prometheus.NewCounter(prometheus.CounterOpts{
+				Name: "my_collector_d",
+				ConstLabels: prometheus.Labels{
+					"name": name,
+					"type": "test",
+				},
+			}),
+		}
+
+		// Register should not fail, since each collector has a unique
+		// set of sub-collectors, determined by their names and const label values.
+		if err := reg.Register(&collector); err != nil {
+			alreadyRegErr, ok := err.(prometheus.AlreadyRegisteredError)
+			if !ok {
+				t.Fatal(err)
+			}
+
+			previous := alreadyRegErr.ExistingCollector.(*collidingCollector)
+			current := alreadyRegErr.NewCollector.(*collidingCollector)
+
+			t.Errorf("Unexpected registration error: %q\nprevious collector: %s (i=%d)\ncurrent collector %s (i=%d)", alreadyRegErr, previous.name, previous.i, current.name, current.i)
+		}
 	}
 }
